from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import os
import base64
from enum import Enum

class StorageFormat(Enum):
    BSON = 2
    JSON = 1

# CryptographyDependencies format
# salt_size (4 bytes) | salt (salt_size bytes) | storage_format (4 byte) 
class DefaultCryptography:

    SALT_BYTES = 4
    STORAGE_BYTES = 4
    SALT_LENGTH = 64

    def __init__(self, password, dependencies):
        self.salt_length = int.from_bytes(dependencies[:self.SALT_BYTES], byteorder='big')
        self.salt = dependencies[self.SALT_BYTES:self.salt_length+self.SALT_BYTES]
        self.storage_format = StorageFormat(int.from_bytes(dependencies[self.salt_length+self.SALT_BYTES:self.salt_length+self.SALT_BYTES+self.STORAGE_BYTES], byteorder='big'))

        self.password = password
        if (type(self.password) == str):
            self.password = self.password.encode("utf-8")
        self.key = self.deriveKey()
        self.f = Fernet(self.key)


    @staticmethod
    def generateCryptographyDependencies(storage_format=StorageFormat.BSON):
        salt = os.urandom(DefaultCryptography.SALT_LENGTH)
        salt_length_bytes = DefaultCryptography.SALT_LENGTH.to_bytes(DefaultCryptography.SALT_BYTES, byteorder='big')
        return salt_length_bytes + salt + storage_format.value.to_bytes(DefaultCryptography.STORAGE_BYTES, byteorder='big')
    
    def deriveKey(self):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=self.salt,
            iterations=100000,
            backend=default_backend()
        )
        key = kdf.derive(self.password)
        return base64.urlsafe_b64encode(key)

    def encrypt(self, raw_data):
        if (type(raw_data) == str):
            raw_data = raw_data.encode("utf-8")
        return self.f.encrypt(raw_data)

    def decrypt(self, encrypted_data):
        return self.f.decrypt(encrypted_data)

    def hash(self, raw_data):
        return hashes.SHA256(raw_data)


import random
import string

def generateKey(length, uppers=True, lowers=True, digits=True, punctuation=True, except_chars=""):
    pool = ""
    if (uppers):
        pool += string.ascii_uppercase
    if (lowers):
        pool += string.ascii_lowercase
    if (digits):
        pool += string.digits
    if (punctuation):
        pool += string.punctuation
    for c in except_chars:
        pool = pool.replace(c, "")
    return ''.join(random.SystemRandom().choice(pool) for _ in range(length))



from tinydb.storages import Storage
from security import DefaultCryptography, StorageFormat
import os
import json
import bson

# File header
# header_length (4 bytes) | crypt_dependencies (header_length bytes)

class SecureStorage(Storage):

    def __init__(self, filename, password):
        if (os.path.isfile(filename)):
            self.file = open(filename, "rb+")


            self.header_length = int.from_bytes(self.file.read(4), byteorder='big')
            crypt_dependencies = self.file.read(self.header_length)


            self.cript = DefaultCryptography(password, crypt_dependencies)

            #Try to decrypt
            self.file.seek(4+self.header_length)

            if (self.cript.storage_format == StorageFormat.JSON):
                a = self.file.read()
                self.cript.decrypt(a).decode("utf-8")
            else:
                self.cript.decrypt(self.file.read())

        else:
            raise FileNotFoundError("Database file not found")

    @staticmethod
    def create(filename, password, storage_format=StorageFormat.JSON):
        if (os.path.isfile(filename)):
            raise Exception("Database file already exists")
        else:
            with open(filename, 'wb+') as file:
                crypt_dependencies = DefaultCryptography.generateCryptographyDependencies(storage_format=storage_format)
                cript = DefaultCryptography(password, crypt_dependencies)

                length_bytes = len(crypt_dependencies).to_bytes(4, byteorder='big') #4 bytes
                file.write(length_bytes)
                file.write(crypt_dependencies)
                
                if (storage_format == StorageFormat.BSON):   
                    data = cript.encrypt(bson.dumps({}))
                else:
                    data = cript.encrypt('{}')
                file.write(data)

    
    def read(self):
        self.file.seek(4+self.header_length)
        
        if (self.cript.storage_format == StorageFormat.BSON): 
            data = self.cript.decrypt(self.file.read())
            return bson.loads(data)
        else:
            data = self.cript.decrypt(self.file.read()).decode("utf-8")
            return json.loads(data)

    def write(self, data):
        self.file.seek(4+self.header_length)
        if (self.cript.storage_format == StorageFormat.BSON):      
            crypted = self.cript.encrypt(bson.dumps(data))
        else:
            crypted = self.cript.encrypt(json.dumps(data).encode("utf-8"))
        self.file.write(crypted)
        self.file.flush()
        os.fsync(self.file.fileno())
        self.file.truncate()

    def close(self):
        self.file.close()
        


#!/usr/bin/python3 -O

from argparse import ArgumentParser
from tinydb import TinyDB, Query, operations
from getpass import getpass
from datetime import datetime
from cryptography.fernet import InvalidToken
from security import DefaultCryptography, StorageFormat
from storage import SecureStorage
from utils import generateKey
import os
import sys
import pyperclip

def create(args):
    format = args.storage_format.lower()
    if (format == 'json'):
        storage_format = StorageFormat.JSON
    elif (format == 'bson'):
        storage_format = StorageFormat.BSON
    else:
        storage_format = None
        print("Storage format not exists", file=sys.stderr)
        exit(-1)

    print("Creating new database in", args.file, "using",storage_format.name, "format")

    if (os.path.isfile(args.file)):
        print("Database file already exists", file=sys.stderr)
        exit(-1)
    else:
        password = getpass(prompt="Insert your master password: ")
        conf_password = getpass(prompt="Confirm your master password: ")

        if (password != conf_password):
            print("Passwords do not match", file=sys.stderr)
            exit(-1)
        else:
            try:
                SecureStorage.create(filename=args.file, password=password, storage_format=storage_format)
                print("Database created successfully")
            except Exception as err:
                print("There something wrong", file=sys.stderr)
                print(err, file=sys.stderr)
                exit(-1)

def login(database_name):
    if (not os.path.isfile(database_name)):
        print("Database not exists", file=sys.stderr)
        exit(-1)
    password = getpass(prompt="Password for " + database_name + ": ")

    try:
        db = TinyDB(args.file, password, storage=SecureStorage)
        return db
    except InvalidToken:
        print("Wrong password", file=sys.stderr)
        exit(-1)

def add(args):
    print("Adding password labeled ", args.label, " to the database in ", args.file)

    db = login(args.file)

    query = Query()
    if (len(db.search(query.label == args.label)) > 0):
        print("This label already in use")
        exit(-1)
    new_password = getpass(prompt="Insert password for " + args.label+ ": ")
    conf_password = getpass(prompt="Confirm your password: ")

    if (new_password != conf_password):
        print("Passwords don't match")
        exit(-1)
    db.insert({'label': args.label, 'password': new_password, 'created_date': datetime.now().isoformat(), 'last_access': None, 'last_update': None})

    print("Successfully")

def get(args):

    db = login(args.file)

    query = Query()
    result = db.search(query.label == args.label)
    if (len(result) == 0):
        print("This label not exists in this database")
        exit(-1)


    password_model = result[0]
    password = password_model['password']
    

    if (password_model["last_access"] != None):
        last_access = datetime.strptime(password_model["last_access"], "%Y-%m-%dT%H:%M:%S.%f")
        print("The last access was in ", last_access.strftime("%d/%m/%Y %I:%M%p"))
    else: 
        print("First access for this password")
    
    if (password_model["last_update"] != None):
        last_update = datetime.strptime(password_model["last_update"], "%Y-%m-%dT%H:%M:%S.%f")
        print("The last access was in ", last_update.strftime("%d/%m/%Y %I:%M%p"))

    if (args.to_clipboard):
        pyperclip.copy(password)
        print("Password copied to clipboard")
    else:
        print("The password for ", args.label, " is: ", password)

    db.update(operations.set("last_access", datetime.now().isoformat()), doc_ids=[password_model.doc_id])

def update(args):
    print("Updating password for ", args.label)
    db = login(args.file)

    query = Query()
    result = db.search(query.label == args.label)
    if (len(result) == 0):
        print("This label not exists in this database")
        exit(-1)

    password_model = result[0]

    new_password = getpass(prompt="Insert a new password for " + args.label+ ": ")
    conf_password = getpass(prompt="Confirm your password: ")

    if (new_password != conf_password):
        print("Passwords don't match")
        exit(-1)

    db.update({"last_update": datetime.now().isoformat(), "password": new_password}, doc_ids=[password_model.doc_id])

    print("Successfully")
    

def remove(args):
    print("Removing password for ", args.label)
    db = login(args.file)


    query = Query()
    result = db.search(query.label == args.label)
    if (len(result) == 0):
        print("This label not exists in this database")
        exit(-1)

    db.remove(doc_ids=[result[0].doc_id])
    print("Password ", args.label, " removed successfully")

def generate(args):
    print("Generating password")
    password = generateKey(length=args.length, uppers=not args.no_uppercase, lowers=not args.no_lowercase, punctuation=not args.no_punctuation, digits= not args.no_digits, except_chars=args.except_chars)
    if (args.save):
        db = login(args.file)
        query = Query()
        if (len(db.search(query.label == args.save)) > 0):
            print("This label already in use")
            exit(-1)
        db.insert({'label': args.save, 'password': password, 'created_date': datetime.now().isoformat(), 'last_access': None, 'last_update': None})
        print("Password has been saved with the label ", args.save)
    if (args.to_clipboard):
        pyperclip.copy(password)
        print("Password copied to clipboard")
    else:
        print("The generated password is: ", password)

def list_labels(args):
    print("Listing all labels")
    db = login(args.file)
    for p in db.all():
        print(p['label'])
    
# Menu 
parser = ArgumentParser()
operation_subparser = parser.add_subparsers(title="Operations", dest="operation")
operation_subparser.required = True
parser.add_argument("-f", "--file", metavar="FILE", help="Database file", default="keyring.db")

create_parser = operation_subparser.add_parser("create", help="Create database keyring")
create_parser.add_argument("-F", "--storage-format", metavar="format", default="bson", help="The storage format (JSON or BSON) [default=BSON]")
create_parser.set_defaults(func=create)

add_parser = operation_subparser.add_parser("add", help="Add a password")
add_parser.add_argument("label", help="label for the password")
add_parser.set_defaults(func=add)

get_parser = operation_subparser.add_parser("get", help="get a password")
get_parser.add_argument("label", help="label for the password")
get_parser.add_argument("-c", "--to-clipboard", action="store_true", help="Copy the password to clipboard and don't show it")
get_parser.set_defaults(func=get)

update_parser = operation_subparser.add_parser("update", help="update a password")
update_parser.add_argument("label", help="label for the password")
update_parser.set_defaults(func=update)

remove_parser = operation_subparser.add_parser("remove", help="Remove a password")
remove_parser.add_argument("label", help="label for the password to remove")
remove_parser.set_defaults(func=remove)

generate_parser = operation_subparser.add_parser("generate", help="generate a password")
generate_parser.add_argument("-l", "--length", metavar="length", type=int, default="12", help="The length for the generated password [default=12]")
generate_parser.add_argument("-u", "--no-uppercase", action="store_true", help="Don't use uppercase chars in the password")
generate_parser.add_argument("-ll", "--no-lowercase", action="store_true", help="Don't use lowercase chars in the password")
generate_parser.add_argument("-d", "--no-digits", action="store_true", help="Don't use digits in the password")
generate_parser.add_argument("-p", "--no-punctuation", action="store_true", help="Don't use punctuation chars in the password")
generate_parser.add_argument("-e", "--except-chars", metavar="except_chars", default="", help="Don't use these chars in the password")
generate_parser.add_argument("-s", "--save", metavar="label", help="Save the generate password using this label")
generate_parser.add_argument("-c", "--to-clipboard", action="store_true", help="Copy the password to clipboard and don't show it")
generate_parser.set_defaults(func=generate)

list_parser = operation_subparser.add_parser("list", help="List all password labels")
list_parser.set_defaults(func=list_labels)


if __name__ == "__main__":
    args = parser.parse_args()
    args.func(args)

    



# coding=utf-8
from setuptools import setup, find_packages
import os

def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    return open(path, encoding='utf-8').read()

if os.name == "nt":
    scripts = None
    entry_points = {
        'console_scripts': [
            'keyring.py=keyring:_main'
        ],
    }
else:
    scripts = ['keyring.py']
entry_points = None

setup(
    name='pyKeyring',
    py_modules=['keyring', 'security', 'storage', 'utils'],
    version="0.6",
    description='A simple and secure tool to store passwords',
    long_description=read('README.rst'),
    url='https://github.com/gabrielperes97/pyKeyring',
    author='Gabriel Leopoldino',
    author_email='gabrielperes97@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3',
	    'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.4',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: Utilities'
    ],
    tests_require=[],
    install_requires=[
        'tinydb>=3.11.1',
        'cryptography==3.3.2',
        'pyperclip>=1.6.4',
        'bson>=0.5.6'
    ],
    scripts=scripts,
    entry_points=entry_points,
)
